/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/***************************************************************************
 *            mov_encoder_thread.h
 *
 *  Tue May 17 16:00:01 CEST 2005
 *  Copyright  2005 Bent Bisballe
 *  deva@aasimon.org
 ****************************************************************************/

/*
 *    This file is part of MIaV.
 *
 *    MIaV is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    MIaV is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with MIaV; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
#include "config.h"
#ifndef __MIAV_MOV_ENCODER_THREAD_H__
#define __MIAV_MOV_ENCODER_THREAD_H__

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <vector>
using namespace std;

#include "mov_encoder.h"
#include "audio_encoder.h"
#include "mov_encoder_writer.h"

#include "info.h"

class MovEncoderThread {
public:
  MovEncoderThread(const char *cpr, Info *info);
  ~MovEncoderThread();

  void encode(Frame* frame);

private:
  Info *info;

  FrameVectorQueue *inputqueue;
  FramePriorityQueue *video_outputqueue;
  FramePriorityQueue *audio_inputqueue;
  FramePriorityQueue *audio_outputqueue;
  FrameVector *block;

  //thread stuff
  sem_t in_sem;
  sem_t video_out_sem;
  sem_t audio_in_sem;
  sem_t audio_out_sem;

  sem_t read_sem;

  pthread_mutex_t input_mutex;
  pthread_mutex_t video_output_mutex;
  pthread_mutex_t audio_input_mutex;
  pthread_mutex_t audio_output_mutex;

  volatile bool movencodersrunning;

  // Used for encoder switching
  unsigned int frame_number;

  unsigned int num_frames_in_block;

  MovEncoderWriter *writer;
  //  pthread_t* writer_tid;

  AudioEncoder* audioenc;

  int threads;
  vector<MovEncoder*> encs;
  //  vector<pthread_t*> tids;
};

#endif/*__MIAV_MOV_ENCODER_THREAD_H__*/
